import asyncio
import pandas as pd
from dictator import Dictator

class DictatorTemperatureExperiment:
    debug = True

    def __init__(self, models: list[str], temperatures: list[float], amount: int, iterations: int, output_file: str):
        self.models = models
        self.temperatures = temperatures
        self.amount = amount
        self.iterations = iterations
        self.output_file = output_file  # Path to the CSV output file

    async def run_experiment(self):
        results = []
        hallucinations = {model: 0 for model in self.models}  # Track hallucinations for each model
        # Loop through different temperatures
        for temp in self.temperatures:
            # Loop through different models
            for model in self.models:
                if self.debug:
                    print(f"Running experiment for model: {model} at temperature: {temp}")
                # Loop through iterations
                for iteration in range(1, self.iterations + 1):
                    valid_response = False
                    while not valid_response:
                        game_agent = Dictator(amount=self.amount, model=model, temperature=temp)
                        response = await game_agent.run()

                        # Check if the shares are within the valid range [0, 100]
                        if 0 <= response['my_share'] <= 100 and 0 <= response['other_share'] <= 100:
                            valid_response = True
                        else:
                            # If hallucination, increment the hallucination counter and rerun the experiment for this iteration
                            hallucinations[model] += 1
                            if self.debug:
                                print(f"Hallucination detected for model: {model} at iteration {iteration}. Re-running...")

                    # Debugging output
                    if self.debug:
                        print(f"Iteration {iteration} - Temp: {temp}, Model: {model}, My Share: {response['my_share']}, Other Share: {response['other_share']}")

                    # Append results as a single row
                    results.append({
                        'Iteration': iteration,
                        'Model': model,
                        'Temperature': temp,
                        'Amount': self.amount,
                        'My Share': response['my_share'],
                        'Other Share': response['other_share']
                    })
                    # Convert to DataFrame and save to CSV after each iteration
                    df = pd.DataFrame(results)
                    df.to_csv(self.output_file, index=False)

        # Print hallucination count for each model at the end of the experiment
        for model, count in hallucinations.items():
            print(f"Model: {model} had {count} hallucinations.")

        return df


# Running the experiment
if __name__ == "__main__":
    models = ["gpt-4.5-preview-2025-02-27", "llama3", "mistral-small", "deepseek-r1"]
    temperatures = [round(i * 0.1, 1) for i in range(7,11)]  # Temperatures from 0.0 to 1.0 with a step of 0.1
    amount = 100
    iterations = 30
    output_file = '../../data/dictator/dictator_temperature.csv'
    experiment = DictatorTemperatureExperiment(models=models, temperatures=temperatures, amount=amount, iterations=iterations, output_file=output_file)
    df = asyncio.run(experiment.run_experiment())
    print(f"Experiment results saved to {output_file}")
    # Model: gpt-4.5 had no hallucination.
    # Model: llama3 had 1 hallucinations.
    # Model: mistral-small had 2 hallucinations.
    # Model: deepseek-r1 had 2 hallucinations.