import asyncio
import os
import random
from typing import Dict, List
import numpy as np
from autogen_agentchat.agents import AssistantAgent
from autogen_agentchat.messages import TextMessage
from autogen_core import CancellationToken
from autogen_ext.models.openai import OpenAIChatCompletionClient
from pydantic import BaseModel
from scipy.optimize import linprog
import requests
import json
import httpx
import re

# Load API key from environment variables
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
if not OPENAI_API_KEY:
    raise ValueError("Missing OPENAI_API_KEY. Set it as an environment variable.")
PAGODA_API_KEY = os.getenv("PAGODA_API_KEY")
if not PAGODA_API_KEY:
    raise ValueError("Missing PAGODA_API_KEY. Set it as an environment variable.")

# Define the expected response format as a Pydantic model
class AgentResponse(BaseModel):
    assetA: float
    assetB: float
    reasoning: str

# The investment game simulation class
class Investment:
    def __init__(self, model: str, temperature: float, max_retries: int = 3):
        self.debug = True
        self.model = model
        self.temperature = temperature
        self.strategy = random
        self.max_retries = max_retries  # Maximum retry attempts in case of hallucinations

        is_openai_model = model.startswith("gpt")
        is_pagoda_model = ":" in model

        base_url = (
            "https://api.openai.com/v1" if is_openai_model else
            "https://ollama-ui.pagoda.liris.cnrs.fr/ollama/api/generate" if is_pagoda_model else
            "http://localhost:11434/v1"
        )
        self.base_url = base_url

        key = OPENAI_API_KEY if is_openai_model else PAGODA_API_KEY

        if not model in ["random", "optimal"]:
            model_info = {
                "temperature": self.temperature,
                "function_calling": True,
                "parallel_tool_calls": True,
                "family": "unknown",
                "json_output": True,
                "vision": False
            }

            self.model_client = OpenAIChatCompletionClient(
                model=self.model,
                base_url=base_url,
                api_key=key,
                model_info=model_info,
                response_format=AgentResponse
            )

    async def run(self, m: float, n: float) -> Dict:
        if self.model == "random":
            return self.apply_random(m, n)
        if self.model == "optimal":
            return self.apply_optimal(m, n)

        instruction = f"""
        You have 100 points to allocate between Asset A and Asset B.
        In this round, allocating 1 point to Asset A yields {m} dollars,
        while allocating 1 point to Asset B yields {n} dollars.
        What is your allocation?
        Your response should be in JSON format with `assetA`, `assetB`, and `reasoning`.
        """

        is_pagoda_model = ":" in self.model
        if is_pagoda_model:
            return await self.run_pagoda(instruction)

        for attempt in range(self.max_retries):
            agent = AssistantAgent(
                name="Investor",
                model_client=self.model_client,
                system_message="You are a helpful assistant."
            )

            response = await agent.on_messages(
                [TextMessage(content=instruction, source="user")],
                cancellation_token=CancellationToken(),
            )

            try:
                response_data = response.chat_message.content
                agent_response = AgentResponse.model_validate_json(response_data)
                assetA, assetB = agent_response.assetA, agent_response.assetB
                if self.debug:
                    print(f"Response (Attempt {attempt+1}): {response_data}")
                if 0 <= assetA <= 100 and 0 <= assetB <= 100 and assetA + assetB == 100:
                    return agent_response.model_dump()
            except Exception as e:
                print(f"Error parsing response (Attempt {attempt+1}): {e}")
        raise ValueError("Model failed to provide a valid response after multiple attempts.")

    async def run_pagoda(self, instruction: str) -> Dict:
        """Runs the Pagoda model using a direct request with improved response parsing."""
        url = self.base_url
        headers = {
            "Authorization": f"Bearer {PAGODA_API_KEY}",
            "Content-Type": "application/json"
        }
        payload = {
            "model": self.model,
            "temperature": self.temperature,
            "prompt": instruction,
            "stream": False,
            "response_format": "json"
        }

        # Print equivalent cURL command for debugging
        curl_cmd = f"""
        curl -X POST {url} \\
             -H "Authorization: Bearer {PAGODA_API_KEY}" \\
             -H "Content-Type: application/json" \\
             -d '{json.dumps(payload, indent=2)}'
        """
        print("Run this cURL command in your terminal to manually test the request:\n")
        print(curl_cmd)

        for attempt in range(self.max_retries):
            try:
                async with httpx.AsyncClient(verify=False, timeout=30) as client:
                    response = await client.post(url, headers=headers, json=payload)
                    response.raise_for_status()  # Raise an error for HTTP status codes 4xx/5xx
                    response_data = response.json()

                if self.debug:
                    print(f"Raw response (Attempt {attempt + 1}): {response_data}")

                response_json = response_data.get("response")
                if not response_json:
                    raise ValueError(f"Missing 'response' field (Attempt {attempt + 1})")

                # Print full response for debugging
                if self.debug:
                    print(f"Full response content (Attempt {attempt + 1}): {response_data}")

                # Clean and parse JSON
                if isinstance(response_json, str):
                    response_json = response_json.strip()
                    match = re.search(r'```json\n(.*?)\n```', response_json, re.DOTALL)
                    if match:
                        response_json = match.group(1)

                    response_json = response_json.replace("\n", "").replace("\\", "")

                    try:
                        response_dict = json.loads(response_json)
                    except json.JSONDecodeError:
                        print(f"Failed JSON: {response_json}")
                        raise ValueError(f"Failed to parse JSON (Attempt {attempt + 1})")

                elif isinstance(response_json, dict):
                    response_dict = response_json
                else:
                    raise TypeError(f"Unexpected response type: {type(response_json)}")

                # Validate and adjust asset allocation
                agent_response = AgentResponse(**response_dict)
                assetA, assetB = round(agent_response.assetA), round(agent_response.assetB)
                difference = 100 - (assetA + assetB)
                if assetA >= assetB:
                    assetA += difference
                else:
                    assetB += difference

                if assetA + assetB != 100:
                    raise ValueError(f"Invalid allocation sum: {assetA}, {assetB}")

                if not (0 <= assetA <= 100 and 0 <= assetB <= 100):
                    raise ValueError(f"Invalid asset allocation: {assetA}, {assetB}")

                return agent_response.dict()

            except httpx.HTTPStatusError as e:
                print(
                    f"HTTP error from Pagoda API (Attempt {attempt + 1}): {e.response.status_code} - {e.response.text}")
            except httpx.RequestError as e:
                print(f"Request error in Pagoda API (Attempt {attempt + 1}): {e}")
            except json.JSONDecodeError as e:
                print(f"JSON parsing error (Attempt {attempt + 1}): {e}")
            except ValueError as e:
                print(f"Value error (Attempt {attempt + 1}): {e}")
            except Exception as e:
                print(f"Unexpected error (Attempt {attempt + 1}): {e}")

        raise ValueError("Pagoda model failed to provide a valid response after multiple attempts.")


    async def run_rounds(self, nb_rounds: int) -> float:
        """Runs the investment game for n rounds and computes the CCEI."""
        results = []
        prices = []
        choices = []
        budgets = []
        for _ in range(nb_rounds):
            m, n = self.generate_M_N()
            if self.debug:
                print(f"m, n: {m}, {n}")
            result = await self.run(m, n)
            if self.debug:
                print(f"result: {result}")
            results.append(result)
            prices.append([m, n])
            budget = round(m * result['assetA'] + n * result['assetB'], 4)
            budgets.append(budget)
            choices.append([result['assetA'], result['assetB']])

        # Compute CCEI
        ccei_value = self.compute_ccei(prices, choices, budgets)
        if self.debug:
            print(f"prices: {prices}")
            print(f"choices: {choices}")
            print(f"budgets: {budgets}")
            print(f"CCEI: {ccei_value}")
        return ccei_value


    def apply_random(self, m: int, n: int) -> Dict:
        assetA = random.randint(0, 100)
        assetB = 100 - assetA
        return {"assetA": assetA, "assetB": assetB, "reasoning": "Random choice"}

    def apply_optimal(self, m: int, n: int) -> Dict:
        assetA = 100 if m > n else 0
        assetB = 100 - assetA
        return {"assetA": assetA, "assetB": assetB, "reasoning": "Optimal choice"}

    def generate_M_N(self):
        while True:
            M = random.uniform(0.1, 1)
            N = random.uniform(0.1, 1)
            if max(M, N) >= 0.5:
                return round(M, 1), round(N, 1)

    def compute_ccei(self, prices, choices, budgets):
        """
        Computes the Critical Cost Efficiency Index (CCEI).
        """
        num_rounds = len(prices)
        lambdas = []
        for i in range(num_rounds):
            # Objective: Maximize lambda (minimize -lambda)
            c = [-1]
            # Constraints: p_t * x_s <= lambda * I_t for all rounds s
            A = []
            b = []
            for j in range(num_rounds):
                p_t = prices[i]
                x_s = choices[j]
                I_t = budgets[i]
                A.append([p_t[0] * x_s[0] + p_t[1] * x_s[1]])
                b.append(I_t)
            # Add constraint that lambda is between 0 and 1
            bounds = [(0, 1)]
            # Solve the linear programming problem
            res = linprog(c, A_ub=A, b_ub=b, bounds=bounds, method='highs')
            if res.success:
                lambdas.append(res.x[0])
            else:
                lambdas.append(0)
        return min(lambdas)  # The CCEI is the minimum lambda across all rounds


# Run the async function and return the response
if __name__ == "__main__":
    game_agent = Investment(model="mistral-small", temperature=0.0)
    response = asyncio.run(game_agent.run_rounds(30))
    print(response)
